
/*=============================================================================*\
 * File: gfsmAlgebra.c
 * Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
 * Description: finite state machine library
 *
 * Copyright (c) 2004 Bryan Jurish.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * See file LICENSE for further informations on licensing terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *=============================================================================*/

#include <glib.h>
#include <gfsmAlgebra.h>

/*======================================================================
 * Methods: algebra: utilities
 */

/*--------------------------------------------------------------
 * concat_final_func()
 *  + called for each final @id of @fsm during concat(@fsm,@fsm2)
 *  + during the call @fsm1->root_id should be set to the translated root of @fsm2
 */
gboolean _gfsm_automaton_concat_final_func(gfsmStateId id, gpointer dummy, gfsmAutomaton *fsm)
{
  gfsm_automaton_add_arc(fsm, id, fsm->root_id, gfsmEpsilon, gfsmEpsilon, gfsmNoWeight);
  gfsm_automaton_find_state(fsm,id)->is_final = FALSE;
  return FALSE;
}

/*--------------------------------------------------------------
 * closure_final_func()
 *  + called for each final @id of @fsm during closure(@fsm)
 */
gboolean _gfsm_automaton_closure_final_func(gfsmStateId id, gpointer dummy, gfsmAutomaton *fsm)
{
  if (id != fsm->root_id)
    gfsm_automaton_add_arc(fsm, id, fsm->root_id, gfsmEpsilon, gfsmEpsilon, gfsmNoWeight);
  return FALSE;
}



/*======================================================================
 * Methods: algebra
 */

/*--------------------------------------------------------------
 * closure()
 */
gfsmAutomaton *gfsm_automaton_closure(gfsmAutomaton *fsm, gboolean is_plus)
{
  //-- sanity check(s)
  if (!fsm || fsm->root_id == gfsmNoState) return fsm;

  //-- add epsilon arcs from old final states to translated new root
  g_tree_foreach(fsm->finals, (GTraverseFunc)_gfsm_automaton_closure_final_func, fsm);

  //-- reflexive+transitive or reflexive?
  if (!is_plus)
    gfsm_automaton_set_final_state(fsm,fsm->root_id,TRUE);

  return fsm;
}

/*--------------------------------------------------------------
 * complement()
 */
gfsmAutomaton *gfsm_automaton_complement(gfsmAutomaton *fsm)
{
  gfsmAlphabet *alph = gfsm_range_alphabet_new();
  gfsm_automaton_get_alphabet(fsm,alph);
  gfsm_automaton_complement_full(fsm,alph);
  gfsm_alphabet_free(alph);
  return fsm;
}

/*--------------------------------------------------------------
 * complement_full()
 */
gfsmAutomaton *gfsm_automaton_complement_full(gfsmAutomaton *fsm, gfsmAlphabet *alph)
{
  gfsmStateId sink_id = gfsm_automaton_add_state(fsm);
  gfsm_automaton_complete(fsm,alph,sink_id);
  //...
  g_printerr("gfsm_automaton_complement(): not yet implemented!");
  g_assert_not_reached();
  return fsm;
}

/*--------------------------------------------------------------
 * complete()
 */
gfsmAutomaton *gfsm_automaton_complete(gfsmAutomaton *fsm, gfsmAlphabet *alph, gfsmStateId sink)
{
  gfsmStateId id;

  g_printerr("gfsm_automaton_complete(): not yet implemented!");
  g_assert_not_reached();

  for (id = 0; id < fsm->states->len; id++) {
    gfsmState *s = gfsm_automaton_find_state(fsm,id);
    if (!s || !s->is_valid) continue;
    //-- argh: automaton must be deterministic for this to work!
  }
  return fsm;
}

/*--------------------------------------------------------------
 * concat()
 */
gfsmAutomaton *gfsm_automaton_concat(gfsmAutomaton *fsm1, const gfsmAutomaton *fsm2)
{
  gfsmStateId offset;
  gfsmStateId id2;

  //-- sanity check(s)
  if (!fsm2 || fsm2->root_id == gfsmNoState) return fsm1;

  offset = fsm1->states->len;
  gfsm_automaton_reserve(fsm1, offset + fsm2->states->len);

  //-- add epsilon arcs from old final states to translated new root
  if (fsm1->root_id != gfsmNoState) {
    gfsmStateId root_tmp = fsm1->root_id;
    fsm1->root_id        = fsm2->root_id+offset;
    g_tree_foreach(fsm1->finals, (GTraverseFunc)_gfsm_automaton_concat_final_func, fsm1);
    fsm1->root_id        = root_tmp;
  } else {
    fsm1->root_id = fsm2->root_id + offset;
  }
  gfsm_set_clear(fsm1->finals);

  for (id2 = 0; id2 < fsm2->states->len; id2++) {
    const gfsmState *s2 = gfsm_automaton_find_state_const(fsm2,id2);
    gfsmState       *s1 = gfsm_automaton_find_state(fsm1,id2+offset);
    gfsmArcList     *a;
    gfsm_state_copy(s1,s2);
    for (a = s1->arcs; a != NULL; a = a->next) {
      a->arc.target += offset;
    }

    //-- check for new final states
    if (s2->is_final) gfsm_set_insert(fsm1->finals, (gpointer)(id2+offset));
  }
		 
  return fsm1;
}


/*--------------------------------------------------------------
 * invert()
 */
gfsmAutomaton *gfsm_automaton_invert(gfsmAutomaton *fsm)
{
  gfsmStateId id;
  for (id=0; id < fsm->states->len; id++) {
    gfsmState   *s = (gfsmState*)(fsm->states->data)+id;
    gfsmArcList *a;
    if (!s) continue;
    for (a = s->arcs; a != NULL; a = a->next) {
      gfsmLabelId tmp = a->arc.lower;
      a->arc.lower    = a->arc.upper;
      a->arc.upper    = tmp;
    }
  }
  return fsm;
}

/*--------------------------------------------------------------
 * project()
 */
gfsmAutomaton *gfsm_automaton_project(gfsmAutomaton *fsm, int which)
{
  gfsmStateId id;
  for (id=0; id < fsm->states->len; id++) {
    gfsmState     *s = (gfsmState*)(fsm->states->data)+id;
    gfsmArcList   *a;
    if (!s) continue;
    for (a = s->arcs; a != NULL; a = a->next) {
      if (which==1) a->arc.upper = a->arc.lower;
      else          a->arc.lower = a->arc.upper;
    }
  }
  return fsm;
}

/*--------------------------------------------------------------
 * reverse()
 */
gfsmAutomaton *gfsm_automaton_reverse(gfsmAutomaton *fsm)
{
  gfsmStateId new_root = gfsm_automaton_add_state(fsm);
  gfsmStateId id;
  gfsmState   *s, *ts;
  gfsmArcList *a, *a_next, *a_prev;

  //-- reverse arc directions, assigning reversed arcs 'target' values as 'old_src+new_root'
  for (id = 0; id < new_root; id++) {
    s = gfsm_automaton_find_state(fsm,id);
    if (!s || !s->is_valid) continue;

    //-- check for old final states
    if (s->is_final) {
      s->is_final = FALSE;
      gfsm_set_remove(fsm->finals,((gpointer)id));
      gfsm_automaton_add_arc(fsm, new_root, id, gfsmEpsilon, gfsmEpsilon, gfsm_sr_one(fsm->sr));
    }
    /*
    else if (!s->arcs) {
      //-- auto-prune
      s->is_valid = FALSE;
      continue;
    }
    */

    //-- reverse arcs
    for (a_prev=NULL, a=s->arcs; a != NULL; a=a_next) {
      a_next = a->next;
      if (a->arc.target >= new_root) {
	//-- already moved
	a_prev = a;
	continue; 
      }

      //-- steal arc
      if (a_prev==NULL) s->arcs = a_next;
      else {
	a_prev->next = a_next;
	a_prev = a;
      }

      //-- move arc
      ts = gfsm_automaton_find_state(fsm,a->arc.target);
      a->next  = ts->arcs;
      ts->arcs = a;
      
      //-- flag as reversed
      a->arc.target = id + new_root;
    }
  }

  //-- sanitize
  for (id=0; id < new_root; id++) {
    s = gfsm_automaton_find_state(fsm,id);
    if (!s || !s->is_valid) continue;
    for (a=s->arcs; a != NULL; a=a->next) {
      a->arc.target -= new_root;
    }
  }

  //-- root flop
  gfsm_automaton_set_final_state(fsm,fsm->root_id,TRUE);
  fsm->root_id = new_root;

  return fsm;
}

/*--------------------------------------------------------------
 * union()
 */
gfsmAutomaton *gfsm_automaton_union(gfsmAutomaton *fsm1, const gfsmAutomaton *fsm2)
{
  gfsmStateId offset;
  gfsmStateId id2;

  //-- sanity check
  if (!fsm2 || fsm2->root_id == gfsmNoState) return fsm1;

  offset = fsm1->states->len;
  gfsm_automaton_reserve(fsm1, offset + fsm2->states->len);

  for (id2 = 0; id2 < fsm2->states->len; id2++) {
    const gfsmState *s2 = gfsm_automaton_find_state_const(fsm2,id2);
    gfsmState       *s1 = gfsm_automaton_find_state(fsm1,id2+offset);
    gfsmArcList     *a;
    gfsm_state_copy(s1,s2);
    for (a = s1->arcs; a != NULL; a = a->next) {
      a->arc.target += offset;
    }
    //-- index final states from @fsm2
    if (s2->is_final) gfsm_set_insert(fsm1->finals, (gpointer)(id2+offset));
  }

  //-- add epsilon arc to translated root state
  if (fsm1->root_id == gfsmNoState) {
    fsm1->root_id = offset + fsm2->root_id;
  } else {
    gfsm_automaton_add_arc(fsm1,
			   fsm1->root_id,
			   offset + fsm2->root_id,
			   gfsmEpsilon,
			   gfsmEpsilon,
			   gfsmNoWeight);
  }
	 
  return fsm1;
}
