/* -*- Mode: C -*- */
/*=============================================================================*\
 * File: gfsmArc.h
 * Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
 * Description: finite state machine library: arcs
 *
 * Copyright (c) 2004 Bryan Jurish.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * See file LICENSE for further informations on licensing terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *=============================================================================*/

#ifndef _GFSM_ARC_H
#define _GFSM_ARC_H

#include <gfsmCommon.h>

/// "Heavy" arc structure
typedef struct _gfsmArc {
  gfsmStateId       target;  ///< ID of target node
  gfsmLabelId       lower;   ///< Lower label
  gfsmLabelId       upper;   ///< Upper label
  gfsmWeight        weight;  ///< arc weight
} gfsmArc;

/// "Heavy" arc-list structure
typedef struct _gfsmArcList {
  gfsmArc               arc;     ///< current arc
  struct _gfsmArcList  *next;    ///< pointer to next arc
} gfsmArcList;

/*======================================================================
 * Methods: Arcs: Constructors etc.
 */
/// \name Arcs: Constructors etc.
//@{
/** Create a new arc (empty) */
#define gfsm_arc_new() g_new0(gfsmArc,1)

/** Initialize an arc (generic) */
gfsmArc *gfsm_arc_init(gfsmArc *a,
		       gfsmStateId dst,
		       gfsmLabelId lo,
		       gfsmLabelId hi,
		       gfsmWeight wt);

/** Destroy an arc */
#define gfsm_arc_free(a) g_free(a)
//@}

/*======================================================================
 * Methods: Arc Lists: Constructors etc.
 */
/// \name Arc Lists: Constructors etc.
//@{
/** Create a new arc-list node (empty) */
#define gfsm_arclist_new() g_new0(gfsmArcList,1)

/** Create a new arc-list node (full) */
#define gfsm_arclist_new_full(dst,lo,hi,wt) \
   ((gfsmArcList*)(gfsm_arc_init(g_malloc(sizeof(gfsmArcList)),dst,lo,hi,wt)))

/** Create and return a copy of an existing arc-list */
gfsmArcList *gfsm_arclist_copy(const gfsmArcList *src);

/** Destroy an arc-list node and all subsequent nodes */
void gfsm_arclist_free(gfsmArcList *al);

/** Free one node of an arc-list */
#define gfsm_arclist_free1(al) g_free(al)
//@}

/*======================================================================
 * Methods: Arcs | Arc Lists: Accessors
 */
///\name Arc | ArcList: Accessors
//@{
/** Get target node of an arc -- may be gfsmNoState */
#define gfsm_arc_target(a) ((a)->target)

/** Get lower label of an arc -- may be gfsmNoLabel */
#define gfsm_arc_lower(a) ((a)->lower)

/** Get upper label of an arc -- may be gfsmNoLabel */
#define gfsm_arc_upper(a) ((a)->upper)

/** Get weight of an arc -- may be gfsmNoWeight */
#define gfsm_arc_weight(a) ((a)->weight)

/** Get pointer to next arc -- may be NULL */
#define gfsm_arc_next(a) ((a)->next)

//@}


/*======================================================================
 * Methods: Arc Lists: Utilities
 */
///\name ArcList: Utilities
//@{
/** Get length of an arc-list (linear time) */
guint gfsm_arclist_length(gfsmArcList *al);

/** 3-way comparison on arcs : lower,upper,target,weight */
gint gfsm_arc_compare(gfsmArc *a1, gfsmArc *a2);
//@}


#endif /* _GFSM_ARC_H */
