/*=============================================================================*\
 * File: gfsmAutomaton.c
 * Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
 * Description: finite state machine library: automata
 *
 * Copyright (c) 2004 Bryan Jurish.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * See file LICENSE for further informations on licensing terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *=============================================================================*/

#include <gfsmAutomaton.h>
#include <gfsmUtils.h>
#include <gfsmBitVector.h>
#include <stdlib.h>

/*======================================================================
 * Constants
 */
const guint gfsmAutomatonDefaultSize  = 128;
//const guint gfsmAutomatonDefaultGet   = 128;

const gfsmAutomatonFlags gfsmAutomatonDefaultFlags =
  {
    TRUE, //-- is_transducer:1
    TRUE, //-- is_weighted:1
    0     //-- unused:2
  };

//const gfsmSRType gfsmAutomatonDefaultSRType = gfsmSRTReal;
const gfsmSRType gfsmAutomatonDefaultSRType = gfsmSRTTropical;

/*======================================================================
 * Methods: Constructors etc.
 */
/*--------------------------------------------------------------
 * new_full()
 */
gfsmAutomaton *gfsm_automaton_new_full(gfsmAutomatonFlags flags, gfsmSRType srtype, guint size)
{
  gfsmAutomaton *fsm = (gfsmAutomaton*)g_new(gfsmAutomaton,1);
  fsm->flags         = flags;
  fsm->sr            = gfsm_semiring_new(srtype);
  fsm->states        = g_array_sized_new(FALSE, TRUE, sizeof(gfsmState), size);
  fsm->finals        = gfsm_set_new(gfsm_uint_compare);
  fsm->root_id       = gfsmNoState;
  return fsm;
}

/*--------------------------------------------------------------
 * copy()
 */
gfsmAutomaton *gfsm_automaton_copy(gfsmAutomaton *dst, const gfsmAutomaton *src)
{
  gfsmStateId id;
  gfsm_automaton_clear(dst);
  gfsm_automaton_reserve(dst,src->states->len);

  dst->flags   = src->flags;
  dst->root_id = src->root_id;
  gfsm_set_copy(dst->finals, src->finals);

  //-- copy semiring
  if (dst->sr && dst->sr->type != src->sr->type) {
    gfsm_semiring_free(dst->sr);
    dst->sr = gfsm_semiring_copy(src->sr);
  }

  for (id = 0; id < src->states->len; id++) {
    const gfsmState *src_s = gfsm_automaton_find_state_const(src,id);
          gfsmState *dst_s = gfsm_automaton_find_state(dst,id);
    gfsm_state_copy(dst_s, src_s);
  }

  return dst;
}

/*--------------------------------------------------------------
 * clear()
 */
void gfsm_automaton_clear(gfsmAutomaton *fsm)
{
  guint i;
  if (!fsm) return;
  for (i=0; fsm->states && i < fsm->states->len; i++) {
    gfsmState *st = gfsm_automaton_find_state(fsm,i);
    if (!st || !st->is_valid) continue;
    gfsm_state_clear(st);
  }
  if (fsm->states) g_array_set_size(fsm->states,0);
  if (fsm->finals) gfsm_set_clear(fsm->finals);
  fsm->root_id = gfsmNoState;
}

/*--------------------------------------------------------------
 * free()
 */
void gfsm_automaton_free(gfsmAutomaton *fsm)
{
  if (!fsm) return;
  gfsm_automaton_clear(fsm);
  if (fsm->sr)     gfsm_semiring_free(fsm->sr);
  if (fsm->states) g_array_free(fsm->states,TRUE);
  if (fsm->finals) gfsm_set_free(fsm->finals);
  g_free(fsm);
}

/*======================================================================
 * Methods: Accessors: Automaton
 */

/*--------------------------------------------------------------
 * set_semiring()
 */
/*
gfsmSemiring *gfsm_automaton_set_semiring(gfsmAutomaton *fsm, gfsmSemiring *sr)
{
  if (fsm->sr) gfsm_semiring_free(fsm->sr);
  fsm->sr = gfsm_semiring_copy(sr);
  return sr;
}
*/

/*--------------------------------------------------------------
 * n_arcs_full()
 */
guint gfsm_automaton_n_arcs_full(gfsmAutomaton *fsm,
				 guint *n_lo_epsilon,
				 guint *n_hi_epsilon,
				 guint *n_both_epsilon)
{
  guint i, total=0;

  if (n_lo_epsilon) *n_lo_epsilon = 0;
  if (n_hi_epsilon) *n_hi_epsilon = 0;
  if (n_both_epsilon) *n_both_epsilon = 0;

  for (i=0; i < fsm->states->len; i++) {
    gfsmState  *st = gfsm_automaton_find_state(fsm,i);
    gfsmArcList *a;
    if (!st || !st->is_valid) continue;

    for (a=st->arcs ; a != NULL; a=a->next) {
      ++total;
      if (n_lo_epsilon && a->arc.lower==gfsmEpsilon) {
	++(*n_lo_epsilon);
	if (n_hi_epsilon && a->arc.upper==gfsmEpsilon) {
	  ++(*n_hi_epsilon);
	  if (n_both_epsilon) ++(*n_both_epsilon);
	}
      }
      else if (n_hi_epsilon && a->arc.upper==gfsmEpsilon) {
	++(*n_hi_epsilon);
      }
    }
  }
  return total;
}

/*--------------------------------------------------------------
 * reserve()
 */
void gfsm_automaton_reserve(gfsmAutomaton *fsm, gfsmStateId nstates)
{
  if (nstates != gfsmNoState && nstates >= fsm->states->len) {
    g_array_set_size(fsm->states,nstates);
  }
}

/*--------------------------------------------------------------
 * is_cyclic_node()
 */
gboolean gfsm_automaton_is_cyclic_state(gfsmAutomaton *fsm,
					gfsmStateId id,
					gfsmBitVector *visited,
					gfsmBitVector *completed)
{
  gfsmState   *s;
  gfsmArcList *a;

  if (gfsm_bitvector_get(visited,id)) {
    if (gfsm_bitvector_get(completed,id)) return FALSE;
    return TRUE;
  }

  s = gfsm_automaton_find_state(fsm,id);
  if (!s || !s->is_valid) return FALSE;  //-- invalid states don't count as cyclic

  //-- mark node as visited (& not completed)
  gfsm_bitvector_set(visited,id,1);
  gfsm_bitvector_set(completed,id,0);

  //-- visit outgoing arcs
  for (a=s->arcs; a != NULL; a=a->next) {
    if (gfsm_automaton_is_cyclic_state(fsm,a->arc.target,visited,completed))
      return TRUE;
  }

  //-- mark node as completed
  //gfsm_bitvector_set(visited,id,0);
  gfsm_bitvector_set(completed,id,1);

  return FALSE;
}

/*--------------------------------------------------------------
 * is_cyclic()
 */
gboolean gfsm_automaton_is_cyclic(gfsmAutomaton *fsm)
{
  gfsmBitVector *visited;    //-- records which states we've visited
  gfsmBitVector *completed;  //-- records which states we've completed
  gboolean       rc;         //-- return value

  if (!fsm || fsm->root_id == gfsmNoState) return FALSE;

  visited   = gfsm_bitvector_sized_new(fsm->states->len);
  completed = gfsm_bitvector_sized_new(fsm->states->len);
  gfsm_bitvector_resize(visited,   fsm->states->len);
  gfsm_bitvector_resize(completed, fsm->states->len);

  rc = gfsm_automaton_is_cyclic_state(fsm, fsm->root_id, visited, completed);

  //-- cleanup
  gfsm_bitvector_free(visited);
  gfsm_bitvector_free(completed);

  return rc;
}

/*--------------------------------------------------------------
 * get_alphabet()
 */
gfsmAlphabet *gfsm_automaton_get_alphabet(gfsmAutomaton *fsm, gfsmAlphabet *alph)
{
  gfsmStateId id;
  //-- ensure alphabet
  if (!alph) alph = gfsm_range_alphabet_new();

  for (id=0; id < fsm->states->len; id++) {
    gfsmArcList *a;
    gfsmState *s = gfsm_automaton_find_state(fsm,id);
    if (!s || !s->is_valid) continue;
    for (a=s->arcs; a != NULL; a=a->next) {
      gfsm_alphabet_get_key(alph, a->arc.lower);
      gfsm_alphabet_get_key(alph, a->arc.upper);
    }
  }
  return alph;
}

/*======================================================================
 * Methods: Accessors: Automaton States
 */

/*--------------------------------------------------------------
 * add_state_full()
 */
gfsmStateId gfsm_automaton_add_state_full(gfsmAutomaton *fsm, gfsmStateId id)
{
  gfsmState *st;

  if (id == gfsmNoState)
    id = fsm->states->len;

  if (id >= fsm->states->len)
    gfsm_automaton_reserve(fsm,id+1);
  
  st           = gfsm_automaton_find_state(fsm,id);
  st->is_valid = TRUE;

  return id;
}

/*--------------------------------------------------------------
 * set_final_state
 */
void gfsm_automaton_set_final_state(gfsmAutomaton *fsm, gfsmStateId id, gboolean is_final)
{
  gfsm_state_set_final(gfsm_automaton_get_state(fsm,id),is_final);
  if (is_final) gfsm_set_insert(fsm->finals,(gpointer)(id));
}


/*======================================================================
 * Methods: Accessors: Automaton Arcs
 */

/*--------------------------------------------------------------
 * add_arc()
 */
gfsmArc *gfsm_automaton_add_arc(gfsmAutomaton *fsm,
				gfsmStateId q1,
				gfsmStateId q2,
				gfsmLabelId lo,
				gfsmLabelId hi,
				gfsmWeight  w)
{
  gfsmState *q1s;
  gfsm_automaton_ensure_state(fsm,q2);
  q1s = gfsm_automaton_get_state(fsm,q1);
  return gfsm_automaton_add_arc_ptr(fsm, q1s, gfsm_arclist_new_full(q2,lo,hi,w));
}

/*--------------------------------------------------------------
 * add_arc_ptr()
 */
gfsmArc *gfsm_automaton_add_arc_ptr(gfsmAutomaton *fsm,
				    gfsmState     *sp,
				    gfsmArcList   *alp)
{
  /*
  //-- unsorted
  alp->next = sp->arcs;
  sp->arcs = alp;
  */
  //-- sorted
  gfsmArcList *sa, *sa_prev;
  for (sa_prev=NULL, sa=sp->arcs;
       sa && gfsm_arc_compare((gfsmArc*)alp,(gfsmArc*)sa) < 0;
       sa_prev=sa, sa=sa->next)
    ;

  alp->next = sa;
  if (sa_prev==NULL) sp->arcs  = alp;
  else           sa_prev->next = alp;

  return (gfsmArc*)alp;
}
